
INFO = {
  "dtiplayground": { 
    "version" : "0.5.8b2"
  },
  "dmriplayground": {
    "version" : "0.5.8b2"
  },
  "dmriprep": {
    "version" : "0.5.8b2"
  },
  "dmriatlas": {
    "version" : "0.2.0b1"
  },
  "dmriautotract" : {
    "version" : "0.0.2b1"
  },
  "dmrifiberprofile" : {
    "version" : "0.5.8b2"
  }
}
